package de.platon42.intellij.jupiter

import com.intellij.jarRepository.JarRepositoryManager
import com.intellij.jarRepository.RemoteRepositoryDescription
import com.intellij.openapi.module.Module
import com.intellij.openapi.roots.ContentEntry
import com.intellij.openapi.roots.DependencyScope
import com.intellij.openapi.roots.ModifiableRootModel
import com.intellij.openapi.vfs.LocalFileSystem
import com.intellij.testFramework.IdeaTestUtil
import com.intellij.testFramework.LightProjectDescriptor
import com.intellij.testFramework.PlatformTestUtil
import com.intellij.testFramework.PsiTestUtil
import com.intellij.testFramework.fixtures.DefaultLightProjectDescriptor
import com.intellij.testFramework.fixtures.JavaCodeInsightTestFixture
import com.intellij.testFramework.fixtures.LightJavaCodeInsightFixtureTestCase
import com.intellij.testFramework.rules.TestNameExtension
import org.jetbrains.idea.maven.utils.library.RepositoryLibraryProperties
import org.junit.jupiter.api.extension.AfterEachCallback
import org.junit.jupiter.api.extension.BeforeEachCallback
import org.junit.jupiter.api.extension.ExtensionContext
import org.junit.jupiter.api.extension.RegisterExtension
import java.nio.file.Paths
import java.util.*
import java.util.function.Consumer
import java.util.stream.Collectors
import java.util.stream.Stream


abstract class AbstractJUnit5TestCase {

    @RegisterExtension
    protected val testNameRule = TestNameExtension()

    @RegisterExtension
    protected val edtInterceptorExtension = EdtInterceptorExtension()

    protected fun getTestName(lowercaseFirstLetter: Boolean): String {
        return PlatformTestUtil.getTestName(testNameRule.methodName, lowercaseFirstLetter)
    }

    @RegisterExtension
    private val testCase = object : LightJavaCodeInsightFixtureTestCase(), BeforeEachCallback, AfterEachCallback {

        lateinit var extensionContext: ExtensionContext

        override fun getProjectDescriptor(): LightProjectDescriptor {
            val testJdk = getMethodOrClassAnnotation(TestJdk::class.java) ?: return super.getProjectDescriptor()
            val projectDescriptor: DefaultLightProjectDescriptor = object : DefaultLightProjectDescriptor({ IdeaTestUtil.getMockJdk(testJdk.value.toJavaVersion()) }) {
                override fun configureModule(module: Module, model: ModifiableRootModel, contentEntry: ContentEntry) {
                    super.configureModule(module, model, contentEntry)
                    val localJars = getMethodOrClassAnnotation(AddLocalJarToModule::class.java)
                    if (localJars != null) {
                        localJars.value.forEach {
                            addJarContaining(
                                model,
                                it.java
                            )
                        }
                    }
                    val mavenDependencies = getMethodOrClassAnnotations(AddMavenDependencyToModule::class.java)
                    mavenDependencies.forEach(Consumer { it: AddMavenDependencyToModule ->
                        addFromMaven(
                            model,
                            it.value,
                            it.includeTransitiveDependencies,
                            it.scope
                        )
                    })
                }
            }

            return projectDescriptor
        }

        fun addJarContaining(model: ModifiableRootModel?, clazz: Class<*>) {
            val filename = clazz.getResource(clazz.simpleName + ".class").file
            val jarName = filename.substring(0, filename.indexOf(".jar") + 4).removePrefix("file:")
            val jarPath = Paths.get(jarName)

            val jarFile = LocalFileSystem.getInstance().refreshAndFindFileByPath(jarName)
            myFixture.allowTreeAccessForFile(jarFile!!)
            PsiTestUtil.addLibrary(
                model!!,
                jarPath.fileName.toString().replace(".jar", ""),
                jarPath.parent.toString(),
                jarPath.fileName.toString()
            )
        }

        fun addFromMaven(
            model: ModifiableRootModel, mavenCoordinates: String,
            includeTransitiveDependencies: Boolean, dependencyScope: DependencyScope?
        ) {
            val remoteRepositoryDescriptions = RemoteRepositoryDescription.DEFAULT_REPOSITORIES
            val libraryProperties = RepositoryLibraryProperties(mavenCoordinates, includeTransitiveDependencies)
            val roots =
                JarRepositoryManager.loadDependenciesModal(model.project, libraryProperties, false, false, null, remoteRepositoryDescriptions)
            val tableModel = model.moduleLibraryTable.modifiableModel
            val library = tableModel.createLibrary(mavenCoordinates)
            val libraryModel = library.modifiableModel
            check(!roots.isEmpty()) { String.format("No roots for '%s'", mavenCoordinates) }

            for (root in roots) {
                libraryModel.addRoot(root.file, root.type)
            }

            val libraryOrderEntry = model.findLibraryOrderEntry(library) ?: throw java.lang.IllegalStateException("Unable to find registered library $mavenCoordinates")
            libraryOrderEntry.scope = dependencyScope!!

            libraryModel.commit()
            tableModel.commit()
        }

        override fun getTestDataPath(): String {
            val testDataPath = getMethodOrClassAnnotation(TestDataPath::class.java) ?: return super.getTestDataPath()
            val testDataSubPath = getMethodOrClassAnnotation(TestDataSubPath::class.java) ?: return testDataPath.value
            return Paths.get(testDataPath.value, testDataSubPath.value).toString()
        }

        fun getMyFixture(): JavaCodeInsightTestFixture = myFixture

        private fun <T : Annotation?> getMethodOrClassAnnotation(clazz: Class<T>): T? {
            var annotation = extensionContext.requiredTestMethod.getAnnotation(clazz)
            if (annotation == null) {
                annotation = extensionContext.requiredTestClass.getAnnotation(clazz)
            }
            return annotation
        }

        private fun <T : Annotation?> getMethodOrClassAnnotations(clazz: Class<T>): List<T> {
            return Stream.of(
                extensionContext.requiredTestMethod.getAnnotationsByType(clazz),
                extensionContext.requiredTestClass.getAnnotationsByType(clazz)
            )
                .flatMap { array: Array<T>? -> Arrays.stream(array) }
                .collect(Collectors.toList())
        }

        override fun beforeEach(context: ExtensionContext) {
            extensionContext = context
            setUp()
        }

        override fun afterEach(context: ExtensionContext) {
            tearDown()
        }
    }

    protected val fixture: JavaCodeInsightTestFixture get() = testCase.getMyFixture()
}
